"""
Default settings for cratis project
"""
import os
from collections import OrderedDict
from typing import TypeVar

from cratis.utils import Resolvable
from django.utils.module_loading import import_string


class FeatureNotLoadedYet(Exception):
    pass


T = TypeVar('T')


class AttributeDictRef(object):

    data = None

    def __init__(self, data):
        super().__init__()
        self.data = data

    def __getattr__(self, name):
        try:
            return self.data[name]
        except KeyError:
            raise AttributeError

    def __setattr__(self, name, value):
        if not self.data:
            return super().__setattr__(name, value)
        self.data[name] = value


class CratisApplication(object):
    """
    App is intended to be used only for
    application introspecion needs.
    """

    def __init__(self, scope):
        super().__init__()

        self.scope = AttributeDictRef(scope)

        self.loaded_features = []
        self.features = []
        self.feature_map = {}

    def setup(self, env='Dev'):
        self.scope.APP = self

        conf_class = self.scope.data[env]

        # Setup settings variables
        config = conf_class()
        props = {x: getattr(config, x) for x in dir(config) if not x.startswith('__')}

        self.scope.data.update(props)

        # prepare feature list to load
        feature_list = self.collect_features_with_deps(self.scope.FEATURES)

        for _feature in feature_list:
            self.load_feature(_feature)

        # resolve variable real values
        for key, val in self.scope.data.items():
            if isinstance(val, Resolvable):
                self.scope.data[key] = val.resolve()

        for _feature in self.features:
            _feature.on_after_load()

        for _feature in self.features:
            _feature.on_startup()

    def load_feature(self, feature):
        feature.app = self
        feature.settings = self.scope
        feature.on_load(self.loaded_features)
        self.loaded_features.append(type(feature))
        self.features.append(feature)
        self.feature_map[feature.__class__] = feature

    def collect_features_with_deps(self, features):
        feature_map = OrderedDict()

        for feature in features:
            self.collect_feature_deps(feature, feature_map)

            # override
            feature_map[feature.__class__] = feature

        all_features = feature_map.values()
        return all_features

    def collect_feature_deps(self, feature, feature_map):
        for requirement in feature.get_required():
            if requirement.__class__ not in feature_map:
                self.collect_feature_deps(requirement, feature_map)
                feature_map[requirement.__class__] = requirement

    def get(self, feature_class: T) -> T:
        if isinstance(feature_class, str):
            feature_class = import_string(feature_class)

        try:
            return self.feature_map[feature_class]
        except KeyError:
            raise FeatureNotLoadedYet(
                'Feature %s is not yet loaded. Did you forget to add it as a requirement?' % feature_class)

    def is_loaded(self, feature_class):
        return feature_class in self.feature_map

    def find_features(self, interface):
        features = []
        for feature in self.features:
            if isinstance(feature, interface):
                features.append(feature)

        return features


class CratisConfig(object):
    """
    Base class for Configurations that will be used with cratis.

    Also class defines ROOT_URLCONF settings variable, that use cratis.urls
    to load all urls collected from your Features. Usually you wouldn't change this
    when working with cratis.

    BASE_DIR just points to project directory, so you can specify paths relatively, ex:
    """

    BASE_DIR = os.environ.get('CRATIS_APP_PATH', '.')
    ROOT_URLCONF = 'cratis.urls'

    FEATURES = ()


def override_feature(features, feature_to_override):
    """
    Replaces feature with the given one.

    :param feature_to_override:
    :return:
    """
    return tuple([x if not isinstance(x, type(feature_to_override)) else feature_to_override for x in features])


def skip_feature(features, feature):
    """
    Remove feature from feature list

    :param feature:
    :return:
    """
    return tuple([x for x in features if not isinstance(x, type(feature))])





def test_wsgi():
    from cratis.wsgi import application

def test_urls():
    from cratis.urls import urlpatterns

def test_django():
    from cratis.django import manage_command
    manage_command(['django', 'check'])


import os
from subprocess import check_output

from cratis.bootstrap import init_app, run_env_command
from tests._markers import slow


@slow
def test_init_in_empty_dir(tmpdir):
    """
    Test check simple operations like open file by path,
    load and save on existing file.

    :param tmpdir:
    :return:
    """

    with tmpdir.as_cwd():
        init_app(cratis_path=os.path.dirname(os.path.dirname(__file__)))

        assert tmpdir.join('settings.py').exists()
        assert tmpdir.join('.pyvenv').exists()

        out = check_output(['.pyvenv/bin/django-manage', 'check'])

        # make sure application is loading
        assert b'System check identified no issues' in out

@slow
def test_init_with_settings_and_repo_package(tmpdir):

    with tmpdir.as_cwd():
        with tmpdir.join('settings.py').open('w') as f:
            f.write(
"""
from cratis.settings import CratisConfig
from features import HelloFeature


class Dev(CratisConfig):
    # boo
    DEBUG = True
    SECRET_KEY = '123'
    """)

        init_app()

        # make sure file is not overriden
        assert '# boo' in  tmpdir.join('settings.py').read()






import _ast
import ast
import os
import re
from _ast import AST


def dump_ast(node, annotate_fields=True, include_attributes=False, indent='  '):  # pragma: no cover
    """
    Return a formatted dump of the tree in *node*.  This is mainly useful for
    debugging purposes.  The returned string will show the names and the values
    for fields.  This makes the code impossible to evaluate, so if evaluation is
    wanted *annotate_fields* must be set to False.  Attributes such as line
    numbers and column offsets are not dumped by default.  If this is wanted,
    *include_attributes* can be set to True.
    """
    def _format(node, level=0):
        if isinstance(node, AST):
            fields = [(a, _format(b, level)) for a, b in ast.iter_fields(node)]
            if include_attributes and node._attributes:
                fields.extend([(a, _format(getattr(node, a), level))
                               for a in node._attributes])
            return ''.join([
                node.__class__.__name__,
                '(',
                ', '.join(('%s=%s' % field for field in fields)
                           if annotate_fields else
                           (b for a, b in fields)),
                ')'])
        elif isinstance(node, list):
            lines = ['[']
            lines.extend((indent * (level + 2) + _format(x, level + 2) + ','
                         for x in node))
            if len(lines) > 1:
                lines.append(indent * (level + 1) + ']')
            else:
                lines[-1] += ']'
            return '\n'.join(lines)
        return repr(node)

    if not isinstance(node, AST):
        raise TypeError('expected AST, got %r' % node.__class__.__name__)
    return _format(node)


def indent_text(nr, text):
    inds = (' ' * nr)
    return inds + ('\n' + inds).join(text.splitlines())



def file_add_import(code, module, objects):
    """
    Add import to file.

    Finds last import in file by analyzing AST tree, and inserts new import after last one.

    :param filename:
    :param module:
    :param objects:
    :return:
    """

    source = ast.parse(code)
    objects = list(objects)

    last_import = 0
    for item in source.body:
        if isinstance(item, (_ast.Import, _ast.ImportFrom)):
            last_import = item.lineno

            if isinstance(item, _ast.ImportFrom):
                if item.module == module:
                    for alias in item.names:
                        if alias.name in objects and alias.asname is None:
                            objects.remove(alias.name)

    # everything imported - skipping
    if not objects:
        return code

    new_lines = ('from %s import %s' % (module, ', '.join(objects)),)

    code = insert_lines_to_code(code, last_import, new_lines)

    return code


def insert_lines_to_code(code, at_line, new_lines):
    """
    Insert set of lines into file at desired position

    :param code:
    :param at_line:
    :param new_lines:
    :return:
    """
    lines = code.splitlines()
    code = '\n'.join(tuple(lines[:at_line]) + new_lines + tuple(lines[at_line:]))
    return code


def settings_add_feature(code, package, feature, default_code=None):
    """
    Register a new feature in settings file.

    Anylyze AST tree to find FEATURE = () construction, and use regexp to carefully
    insert new feature, saving original indentation level.

    :param filename:
    :param feature:
    :return:
    """


    if default_code:
        feature_code = indent_text(4, '%s,' % default_code.strip())
    else:
        feature_code = indent_text(4, '%s(),' % feature)

    code = file_add_import(code, package, (feature,))

    source = ast.parse(code)
    for item in source.body:
        if isinstance(item, _ast.ClassDef) and item.name == 'Dev':

            for subitem in item.body:
                if isinstance(subitem, _ast.Assign) \
                        and isinstance(subitem.targets[0], _ast.Name) \
                        and subitem.targets[0].id == 'FEATURES':

                    # Skip Feature registration if it is already added
                    for val in subitem.value.elts:
                        if isinstance(val, _ast.Call) and val.func.id == feature:
                            return code

                    indent = subitem.targets[0].col_offset
                    line = subitem.targets[0].lineno - 1

                    lines = code.splitlines()

                    start = '\n'.join(lines[:line])
                    end = '\n'.join(lines[line:])

                    end = re.sub('^\s+FEATURES\s*=\s*\(',
                                 '\g<0>\n' + indent_text(indent, feature_code),
                                 end, flags=re.MULTILINE)

                    code = start + '\n' + end

                    return code

            # Seems to be there is no FEATURES = () ... add it!

            new_lines = tuple(indent_text(item.col_offset + 4, """FEATURES = (
%s
)
""" % feature_code).splitlines())

            code = insert_lines_to_code(code, item.lineno, new_lines)

            return code

    print('Nothing added -dev class not found')
    return code



def package_to_path(package_name):
    """
    Convert package name to relative directory name

    :param package_name:
    :return:
    """
    return package_name.replace('.', os.path.sep)

def generate_package(package_name, path=None):
    if not path:
        path = os.getcwd()

    if not isinstance(package_name, list):
        parts = package_name.split('.')
    else:
        parts = package_name

    if len(parts) < 2:
        return

    dirname = os.path.join(path, parts[0])

    if not os.path.exists(dirname):
        os.mkdir(dirname)

    if not os.path.exists(os.path.join(dirname, '__init__.py')):
        with open(os.path.join(dirname, '__init__.py'), 'w+') as f:
            f.write('\n\n')

    if len(parts) > 1:
        generate_package(parts[1:], dirname)


import os

from pkg_resources import Requirement, RequirementParseError


def add_to_requirements(package_name, url=None, file_path='requirements.txt'):

    if not url:
        url = package_name

    if os.path.exists(file_path):
        with open(file_path) as f:
            lines = f.read().strip().splitlines()
    else:
        lines = []

    append = True
    new_lines = []
    for line in lines:
        if line.strip() == '':
            new_lines.append(line)
            continue
        try:
            if line.endswith('#egg=%s' % package_name) or Requirement.parse(line).project_name == package_name:
                new_lines.append(url)
                append = False
                continue

            else:
                new_lines.append(line)
        except RequirementParseError:
            new_lines.append(line)

    if append:
        new_lines.append(url)

    with open(file_path, 'w') as f:
        f.write('\n'.join(new_lines))



import pytest
from cratis.utils import Collectable, Resolvable


def test_resolvable_abstract_working():
    with pytest.raises(TypeError):
        Resolvable()

    with pytest.raises(TypeError):
        class MyCls(Resolvable):
            pass

        MyCls()

    class MyCls2(Resolvable):
        def resolve(self):
            pass

    MyCls2()


def test_collectable():
    boo = Collectable()

    assert isinstance(boo, Resolvable)

    boo.append('123')
    boo.append(['321'])

    assert boo.resolve() == ('123', '321')

def test_collectable_non_unique_list():
    boo = Collectable(unique=False, _format=list)

    assert isinstance(boo, Resolvable)

    boo.append('123')
    boo.append('321')
    boo.append('321')

    assert boo.resolve() == ['123', '321', '321']




import inspect
import os
import copy
from contextlib import contextmanager
from typing import TypeVar

from django.conf.urls import url
from django.utils.log import DEFAULT_LOGGING


from cratis.settings import CratisApplication
from cratis.settings import CratisConfig
from cratis.settings import FeatureNotLoadedYet
from cratis.utils import Collectable

T = TypeVar('T')

class FeatureDependencyFailure(Exception):
    pass

#
# @contextmanager
# def feature(feature_cls: T, calee: Feature, optional=False) -> T:
#     """
#     :param feature_cls:
#     :type feature_cls: str | class
#     :param calee:
#     :param optional:
#     :return:
#     """
#
#     try:
#         _feature = CratisApplication.get(feature_cls)
#         _feature.calee = calee
#         yield _feature
#         _feature.calee = None
#     except (FeatureNotLoadedYet, ImportError) as e:
#         if not optional:
#             raise Exception('by %s' % calee, e)

class BaseFeature(object):
    calee = None
    """
    When future is used by other features using `with feature()` syntax,
    this variable get filled with reference to calee.
    """

    """
    Basement for all the features.
    """
    settings = None  # type: CratisConfig
    app = None  # type: CratisApplication

    _static_requirements = ()


    def __init__(self):
        super().__init__()

        self.__requirements = []

    def set_application(self, app):
        """
        Initialized just before configuration process start
        """
        self.app = app
        self.settings = app.scope

    def get_required(self):
        """
        Returns list of classes of features that must be loaded before this feature
        can be loaded
        """
        return self.__requirements

    def require_if_installed(self):
        return self.require(optional=True)

    @contextmanager
    def use(self, feature_cls: T, require=True) -> T:
        """
        :param feature_cls:
        :type feature_cls: str | class
        :param optional:
        :return:
        """
        try:
            _feature = self.app.get(feature_cls)
            _feature.calee = self
            yield _feature
            _feature.calee = None
        except (FeatureNotLoadedYet, ImportError) as e:
            if require:
                raise Exception('by %s' % self, e)

    def get_deps(self):
        """
        Return list of python dependencies that are required by the feature
        :return:
        """
        return ()

    def on_load(self, loaded_features):
        """
        Loads feature. Used by settings class.
        By default, checks requirements and executes configure_settings() method.

        As an argument accepts list of features loaded before current one.
        """

        self.configure_settings()

    def on_after_load(self):
        """
        Called when setting is configured
        """
        pass

    def on_startup(self):
        """
        Last chance to do something after settings are configured, called even later than on_after_load.
        """
        pass

    def configure_settings(self):
        """
        API method.
        Meant to be overridden by subsequent Features.

        Called inside on_load callback.

        DEPRECATED!!!
        """
        self.init()

    def init(self):
        """
        API method.
        Meant to be overridden by subsequent Features.

        Called inside on_load callback.

        DEPRECATED!!!
        """

    def setup(self):
        """
        API method.
        Meant to be overridden by subsequent Features.

        Called inside on_load callback.
        """
        self.init()

    def configure_urls(self, urls):
        """
        API method.
        Meant to be overridden by subsequent Features.

        Called when django imports cratis.url from cratis.urls module.

        As a parameter accepts urlpatterns variable from cratis.urls
        """


def require(*features):
    def require_decorator(cls):
        cls.get_required = lambda x: features
        return cls

    return require_decorator


class Feature(BaseFeature):
    """
    Feature add some concreate functionality to the BaseFeature class.
    """

    # INSTALLED_APPS = Collectable()
    # MIDDLEWARE_CLASSES = Collectable()
    # TEMPLATE_CONTEXT_PROCESSORS = Collectable()
    # TEMPLATE_DIRS = Collectable()
    # STATICFILES_DIRS = Collectable()

    def get_dir(self):
        return os.path.dirname(inspect.getfile(self.__class__))

    def set_default(self, name, value):

        if not hasattr(self.settings, name):
            setattr(self.settings, name, value)

    def append_apps(self, apps):
        """
        Use this in configure_settings, to append new INSTALLED_APPS.
        """

        if isinstance(apps, str):
            apps = (apps,)

        if not hasattr(self.settings, 'INSTALLED_APPS'):
            self.settings.INSTALLED_APPS = ()

        for app in apps:
            if app not in self.settings.INSTALLED_APPS:
                self.settings.INSTALLED_APPS += (app,)

    def append_middleware(self, classes):
        """
        Use this in configure_settings, to append new middleware classes.
        """

        if isinstance(classes, str):
            classes = (classes,)

        if not hasattr(self.settings, 'MIDDLEWARE_CLASSES'):
            self.settings.MIDDLEWARE_CLASSES = ()

        for classname in classes:
            if classname not in self.settings.MIDDLEWARE_CLASSES:
                self.settings.MIDDLEWARE_CLASSES += (classname,)

    def append_template_processor(self, processors):
        """
        Use this in configure_settings, to append new template processors.
        """
        if isinstance(processors, str):
            processors = (processors,)

        if not hasattr(self.settings, 'TEMPLATE_CONTEXT_PROCESSORS'):
            from django.conf.global_settings import TEMPLATE_CONTEXT_PROCESSORS
            self.settings.TEMPLATE_CONTEXT_PROCESSORS = TEMPLATE_CONTEXT_PROCESSORS

        for classname in processors:
            if classname not in self.settings.TEMPLATE_CONTEXT_PROCESSORS:
                self.settings.TEMPLATE_CONTEXT_PROCESSORS += (classname,)

    def append_template_dir(self, dirs):

        if isinstance(dirs, str):
            dirs = (dirs,)

        if not hasattr(self.settings, 'TEMPLATE_DIRS'):
            from django.conf.global_settings import TEMPLATE_DIRS
            self.settings.TEMPLATE_DIRS = TEMPLATE_DIRS

        self.settings.TEMPLATE_DIRS += tuple(dirs)

    def append_asset_dir(self, dirs):
        if isinstance(dirs, str):
            dirs = (dirs,)

        if not hasattr(self.settings, 'STATICFILES_DIRS'):
            from django.conf.global_settings import STATICFILES_DIRS
            self.settings.STATICFILES_DIRS = STATICFILES_DIRS

        self.settings.STATICFILES_DIRS += tuple(dirs)


class Common(Feature):
    """
    This feature is used by most of the django-applications
    """

    def __init__(self, sites_framework=True):
        super().__init__()

        self.sites_framework = sites_framework

    def configure_settings(self):
        s = self.settings

        self.append_apps([
            'django.contrib.auth',
            'django.contrib.contenttypes',
            'django.contrib.sessions',
            'django.contrib.messages',
            'django.contrib.staticfiles',
        ])

        self.append_middleware([
            'django.contrib.sessions.middleware.SessionMiddleware',
            'django.middleware.locale.LocaleMiddleware',
            'django.middleware.common.CommonMiddleware',
            'django.middleware.csrf.CsrfViewMiddleware',
            'django.contrib.auth.middleware.AuthenticationMiddleware',
            'django.contrib.messages.middleware.MessageMiddleware',
            'django.middleware.clickjacking.XFrameOptionsMiddleware',
        ])

        s.STATIC_URL = '/static/'
        s.STATIC_ROOT = s.PROJECT_DIR + '/var/static'

        s.MEDIA_URL = '/media/'
        s.MEDIA_ROOT = s.PROJECT_DIR + '/var/media'

        s.USE_TZ = True

        if self.sites_framework:
            self.append_apps([
                'django.contrib.sites',
            ])
            s.SITE_ID = 1

        s.LOGGING = copy.deepcopy(DEFAULT_LOGGING)

    def configure_urls(self, urls):
        # if self.settings.DEBUG:
        from django.views.static import serve

        def serve_cors(*args, **kwargs):
            response = serve(*args, **kwargs)

            response['Access-Control-Allow-Origin'] = "*"

            return response


        urls += [
            url(r'^media/(?P<path>.*)$', serve_cors, {'document_root': self.settings.MEDIA_ROOT}),
            url(r'^static/(?P<path>.*)$', serve_cors, {'document_root': self.settings.STATIC_ROOT}),
        ]




import pytest
from cratis.add import parse_repo_parts, Url, compile_url, add_cmd
from cratis.bootstrap import init_app, run_env_command
from tests._markers import slow


@pytest.mark.parametrize("url,expected", [
    ("admin", Url(
        feature='cratis-admin',
        user='django-cratis',
        version='master',
        repo='cratis-admin',
        protocol='git+https'
    )),
    ("admin.", Url(
        feature='admin',
        user='django-cratis',
        version='master',
        repo='admin',
        protocol='git+https'
    )),
    ("admin@boo", Url(
        feature='cratis-admin',
        user='boo',
        version='master',
        repo='cratis-admin',
        protocol='git+https'
    )),
    ("admin@boo#dev", Url(
        feature='cratis-admin',
        user='boo',
        version='dev',
        repo='cratis-admin',
        protocol='git+https'
    )),
    ("ssh:baz.@foo/bar#dev", Url(
        feature='baz',
        user='foo',
        version='dev',
        repo='bar',
        protocol='git+ssh'
    )),
])
def test_repo_parts(url, expected):
    result = parse_repo_parts(url)

    assert result == expected


@pytest.mark.parametrize("params, url", [
    (Url(
        feature='baz',
        user='foo',
        version='dev',
        repo='bar',
        protocol='git+ssh'
    ), 'git+ssh://github.com/foo/bar.git@dev#egg=baz'),
])
def test_compile_url(params, url):
    result = compile_url(params)

    assert result == url

@slow
def test_install_admin(tmpdir):

    with tmpdir.as_cwd():
        init_app()

        add_cmd('admin')

        settings = tmpdir.join('settings.py').read()
        assert 'cratis_admin.features import AdminArea' in settings
        assert 'AdminArea(' in settings


import sys

from cratis.generators.module import SettingsModuleFile
from cratis.generators.requirements import add_to_requirements
from pkg_resources import load_entry_point, iter_entry_points

import re

from collections import namedtuple
from cratis.bootstrap import run_env_command


Url = namedtuple('Url', ['feature', 'user', 'version', 'repo', 'protocol'])


# ssh://admin.@django-cratis/cratis-admin#master

def parse_repo_parts(url):

    rx = '^((?P<protocol>[\w-]+):)?(?P<feature>[\w-]+)(?P<exact_feature>\.)?(@(?P<user>[\w-]+))?(/(?P<repo>[\w-]+))?(#(?P<version>[\w-]+))?'

    m = re.match(rx, url)

    feature = m.group('feature')
    exact_feature = m.group('exact_feature')
    if not exact_feature:
        feature = 'cratis-%s' % feature

    user = m.group('user') or 'django-cratis'
    version = m.group('version') or 'master'
    repo = m.group('repo') or feature
    protocol = m.group('protocol') or 'https'
    protocol = 'git+%s' % protocol

    return Url(
        feature=feature,
        user=user,
        version=version,
        repo=repo,
        protocol=protocol
    )


def compile_url(url: Url):
    return '%(protocol)s://github.com/%(user)s/%(repo)s.git@%(version)s#egg=%(feature)s' % url._asdict()


def add_cmd(url, **kwargs):
    data = parse_repo_parts(url)
    url = compile_url(data)

    requirement_url = '-e %s' % url
    run_env_command('pip install %s' % requirement_url)
    run_env_command('cratis-add %s' % data.feature)

    add_to_requirements(data.feature, requirement_url)


def add_feature_to_settings():

    if len(sys.argv) < 2:
        print("Not enough arguments - exit")

    print('\nAdding feature to settings: %s' % sys.argv[1])

    found = []
    for entry_point in iter_entry_points(group='features', name=sys.argv[1]):
        found.append(entry_point.load())

    if not found:
        print("Feature not found")
        sys.exit(1)

    cls = found[0]

    f = SettingsModuleFile('settings')
    f.add_feature(cls.__module__, cls.__name__, getattr(cls, "DEFAULT", None))
    f.save()



import os
import sys

from cratis.env import load_env


def manage_command(args=None):

    if os.environ.get('DEBUG') == '1':
        import pydevd
        pydevd.settrace('localhost', port=33322, stdoutToServer=False, stderrToServer=False, suspend=False)

    load_env()

    from django.core.management import execute_from_command_line

    execute_from_command_line(args or sys.argv)



import os
import string
from random import choice


def create_settings():
    if os.path.exists('settings.py'):
        print('settings.py already exists')
        return

    tpl = """
from cratis.settings import CratisConfig, CratisApplication as App


class Dev(CratisConfig):
    DEBUG = True
    SECRET_KEY = '%s'

    FEATURES = (
        # your features here
    )

App(locals()).setup()
""" % ''.join(choice(string.ascii_letters + string.digits) for _ in range(32))

    with open('settings.py', 'w+') as f:
        f.write(tpl)

    print('Ok. settings.py generated')


def run_env_command(command):
    return os.system('.pyvenv/bin/%s' % command)


def init_app(cratis_path=None, **kwargs):
    if not os.path.exists('.pyvenv'):
        os.system('pyvenv .pyvenv')

    if cratis_path:
        print(cratis_path)
        run_env_command('pip install -e %s[django]' % cratis_path)
    else:
        run_env_command('pip install django-cratis[django]')

    create_settings()


import os

from cratis.generators.codegen import package_to_path, generate_package, file_add_import, settings_add_feature


class ModuleFile(object):
    """
    Represent any python module
    """

    def __init__(self, package, path=None):
        super(ModuleFile, self).__init__()

        self.package = package
        self.path = path or package_to_path(package) + '.py'

        self.content = None

        self.reload()

    @property
    def exists(self):
        """
        Check if file exists on disk

        :return:
        """

        return os.path.exists(self.path)

    def reload(self):
        """
        Reload file source from file

        :return:
        """
        if not self.exists:
            self.content = None
            return

        with open(self.path) as f:
            self.content = f.read()

    def save(self):
        """
        Dump file contents to disk

        :return:
        """

        generate_package(self.package)

        with open(self.path, 'w+') as f:
            f.write(self.content)

    def add_import(self, package, name):
        self.content = file_add_import(self.content, package, (name,))


class SettingsModuleFile(ModuleFile):
    """
    Represents cratis settings module
    """

    def add_feature(self, package, name, default_code=None):
        self.content = settings_add_feature(self.content, package, name, default_code)



import argparse
import os
import sys

from cratis.add import add_cmd
from cratis.bootstrap import init_app
from cratis.django import manage_command


def django_cmd():
    """
    Command that executes django ./manage.py task + loads environment variables from cratis.yml

    Command also can be executed from sub-folders of project.
    """

    manage_command()


def cratis_cmd():
    """

    """
    print("""
┌─┐┬─┐┌─┐┌┬┐┬┌─┐
│  ├┬┘├─┤ │ │└─┐
└─┘┴└─┴ ┴ ┴ ┴└─┘
""")

    parser = argparse.ArgumentParser(prog='cratis')

    subparsers = parser.add_subparsers()

    # install_subparser = subparsers.add_parser('install', help='Install application dependencies')
    # install_subparser.add_argument("--env", type=str, help="Settings Class (Environment)", default="Dev", nargs='?')
    # install_subparser.add_argument("--dump", action="store_true", default=False, help="Only dump, do not install")
    # install_subparser.add_argument("--upgrade", action="store_true", default=False, help="Update dependencies")
    # install_subparser.set_defaults(func=install_feature_dependencies)

    init_subparser = subparsers.add_parser('init', help='Create empty settings file')
    init_subparser.add_argument("--cratis-path", '--cp', help="Local cratis directory to install with -e attribute")
    init_subparser.set_defaults(func=init_app)

    init_subparser = subparsers.add_parser('add', help='Install feature from github')
    init_subparser.add_argument("url", help="Feature url in format [(ssh|https)://][admin][.][@django-cratis][/cratis-admin][#master]")
    init_subparser.set_defaults(func=add_cmd)

    if len(sys.argv) < 2:
        parser.print_help()
        print()
        return

    args = parser.parse_args(sys.argv[1:])

    if args.func:
        args.func(**vars(args))
        print()
    else:
        parser.print_help()
        print()




import os
import sys

import pip


def check_settings_module_or_exit(env=None):
    try:
        return __import__(os.environ['DJANGO_SETTINGS_MODULE'])
    except ImportError as m:
        print('Can not load settings module. Wrong directory?')
        print('-' * 50)
        print('Error: %s' % m)
        print()

        print()

        raise m


def load_env():
    os.environ['DJANGO_SETTINGS_MODULE'] = os.environ.get('DJANGO_SETTINGS_MODULE', 'settings')

    app_path = os.environ.get('CRATIS_APP_PATH', os.getcwd())
    sys.path = [app_path] + sys.path

    check_settings_module_or_exit()



from setuptools import setup, find_packages


setup(
    name='django-cratis',
    version='0.9.1',
    packages=find_packages(),

    url='',
    license='Simplified BSD License',
    author='Alex Rudakov',
    author_email='ribozz@gmail.com',
    description='Django-cratis is a way to group together django applications, so they form reusable features.',
    long_description='',

    extras_require={
        'django':  ["django>1.8.0,<1.10", 'pip>8.0.0'],
        'mysql': ["mysqlclient"],
        'postgres': ["psycopg2"]
    },

    entry_points={
        'console_scripts': [
            'django = cratis.cli:django_cmd',
            'cratis = cratis.cli:cratis_cmd',

            'django-manage = cratis.django:manage_command [django]',
            'cratis-add = cratis.add:add_feature_to_settings [django]'
        ]
    },
)




from abc import ABC, abstractmethod


class Resolvable(ABC):

    @abstractmethod
    def resolve(self):
        """
        Returns set of resolved values
        """


class Collectable(Resolvable):

    def __init__(self, unique=True, _format=tuple):
        self.val = ()
        self.unique = unique
        self._format = _format

    def append(self, vals):
        if isinstance(vals, str):
            vals = (vals,)

        for val in vals:
            if not self.unique or (val not in self.val):
                self.val += (val,)

    def resolve(self):
        return self._format(self.val)
